#ifndef PLOCHY_BEZIERPATCH_H
#define PLOCHY_BEZIERPATCH_H

#include "patch.h"
#include <vector>

class TBezierPatch : public TPatch {
private:
	/// Degree of the patch in U direction
	unsigned int fDegreeU;
	/// Degree of the patch in V direction
	unsigned int fDegreeV;
	/// Array of control points
	std::vector<TVector4d> fVertices;
	/// Array of samples across the patch
	std::vector<TVector4d> fSamples;
	/// Flag storing whether current samples are valid
	bool fValid;

	/// Initializes control points uniformly across [0, 1]x[0, 1]
	void reset();
	/// Computes samples of the patch
	void recompute();
protected:
	/// Called when number of samples per segment changes.
	virtual void samplesPerSegmentChanged(unsigned int oldValue,
		unsigned int newValue);
public:
	TBezierPatch();
	virtual ~TBezierPatch();
	/// Returns coordinates of a control point with specified index.
	virtual TVector4d vertex(TIndex index) const;
	/// Sets coordinates of a control point with specified index.
	virtual void setVertex(TIndex index, TVector4d value);
	/// Returns degree of the patch.
	virtual unsigned int degree(TDirection direction) const;
	/// Sets degree of the patch.
	virtual void setDegree(TDirection direction, unsigned int value);
	/// Returns knot from knot vector.
	virtual double knot(TDirection dir, unsigned int index);
	/// Sets knot in knot vector.
	virtual void setKnot(TDirection dir, unsigned int index, double value);
	/// Returns knot vector size.
	virtual unsigned int knotVectorSize(TDirection dir);
	/// Returns index of a control point nearest to the specified line.
	virtual TIndex getVertexNear(const TVector4d& linePoint,
		const TVector4d& lineVector, double maxRadius = 0.05);
	/// Paints the patch according to the specified styles.
	virtual void paint(TIndex activeVertex = TIndex(),
		TParam highlight = TParam(), TDrawMode drawMode = dmAll);
};

#endif

