//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "mainForm.h"
#include "weightForm.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <cmath>
#include "bezierPatch.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TfMain *fMain;

using std::sin;
using std::cos;

//---------------------------------------------------------------------------

void __fastcall TRenderPanel::Paint() {
	(fCallbackOwner->*fCallback)();
}

__fastcall TRenderPanel::TRenderPanel(TComponent* AOwner, TfMain *CallbackOwner,
	TDrawCallback Callback) : TPanel(AOwner) {

	this->fCallback = Callback;
	this->fCallbackOwner = CallbackOwner;
}
//---------------------------------------------------------------------------

__fastcall TfMain::TfMain(TComponent* Owner)
	: TForm(Owner)
{
}
//---------------------------------------------------------------------------
void __fastcall TfMain::FormCreate(TObject *Sender)
{
	fMouseDown = mbMiddle;

	fCursorXY = (TCursor)100;
	fCursorZ = (TCursor)101;
	Screen->Cursors[fCursorXY] = LoadCursor(HInstance, "CURSORXY");
	Screen->Cursors[fCursorZ] = LoadCursor(HInstance, "CURSORZ");

	fRenderPanel = new TRenderPanel(this, this, &TfMain::drawSurface);
	fRenderPanel->Parent = this;
	fRenderPanel->Align = alClient;
	fRenderPanel->ParentColor = false;
	fRenderPanel->Color = clBlack;
	fRenderPanel->OnMouseDown = RenderPanelMouseDown;
	fRenderPanel->OnMouseMove = RenderPanelMouseMove;
	fRenderPanel->OnMouseUp = RenderPanelMouseUp;
	fRenderPanel->OnResize = RenderPanelResize;

	initializeOpenGl();

	glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

	fAzimuth = 20;
	fElevation = 30;
	fZoom = 1;

	glViewport(0, 0, fRenderPanel->Width, fRenderPanel->Height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, 1, 0, 1, -50, 50);
	glMatrixMode(GL_MODELVIEW);
	updateViewMatrix();

	fPatch = new TBezierPatch();

	updateControls();

	iHighlight->Canvas->Brush->Color = clBlack;
	iHighlight->Canvas->Brush->Style = bsSolid;
	iHighlight->Canvas->FillRect(
		TRect(0, 0, iHighlight->Width, iHighlight->Height));

}
//---------------------------------------------------------------------------

void __fastcall TfMain::FormDestroy(TObject *Sender)
{
	if (fRendCont != 0) {
		if (wglMakeCurrent(fDevCont, 0))
			wglDeleteContext(fRendCont);
		fRendCont = 0;
	}

	fRenderPanel->Free();
	delete fPatch;
}
//---------------------------------------------------------------------------

void TfMain::initializeOpenGl() {
	PIXELFORMATDESCRIPTOR pixelFDesc;
	int pixelFormat;
	GLfloat farba[4];

	fDevCont = GetDC(fRenderPanel->Handle);
	if (fDevCont == 0)
		throw Exception("Fatal error - failed to obtain device context!");

	pixelFDesc.nSize = sizeof(PIXELFORMATDESCRIPTOR);
	pixelFDesc.nVersion = 1;
	pixelFDesc.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL |
		PFD_DOUBLEBUFFER;
	pixelFDesc.iPixelType = PFD_TYPE_RGBA;
	pixelFDesc.cColorBits = 32;
	pixelFDesc.cRedBits = 0;
	pixelFDesc.cRedShift = 0;
	pixelFDesc.cGreenBits = 0;
	pixelFDesc.cGreenShift = 0;
	pixelFDesc.cBlueBits = 0;
	pixelFDesc.cBlueShift = 0;
	pixelFDesc.cAlphaBits = 0;
	pixelFDesc.cAlphaShift = 0;
	pixelFDesc.cAccumBits = 0;
	pixelFDesc.cAccumRedBits = 0;
	pixelFDesc.cAccumGreenBits = 0;
	pixelFDesc.cAccumBlueBits = 0;
	pixelFDesc.cAccumAlphaBits = 0;
	pixelFDesc.cDepthBits = 16;
	pixelFDesc.cStencilBits = 0;
	pixelFDesc.cAuxBuffers = 0;
	pixelFDesc.iLayerType = PFD_MAIN_PLANE;
	pixelFDesc.bReserved = 0;
	pixelFDesc.dwLayerMask = 0;
	pixelFDesc.dwVisibleMask = 0;
	pixelFDesc.dwDamageMask = 0;

	pixelFormat = ChoosePixelFormat(fDevCont, &pixelFDesc);
	if (pixelFormat == 0)
		throw Exception("Fatal error - cannot find suitable pixel format!");

	if (!SetPixelFormat(fDevCont, pixelFormat, &pixelFDesc))
		throw Exception("Fatal error - error setting pixel format!");

	fRendCont = wglCreateContext(fDevCont);
	if (fRendCont == 0)
		throw Exception("Fatal error - error creating OpenGL context!");

	if (!wglMakeCurrent(fDevCont, fRendCont))
		throw Exception("Fatal error - error setting OpenGL context!");

	glClearColor(0, 0, 0, 1);
	glClearDepth(1);
	glDepthFunc(GL_LEQUAL);
	glPointSize(3);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_POINT_SMOOTH);
	glEnable(GL_NORMALIZE);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	gluLookAt(1, 0, 0, 0, 0, 0, 0, 0, 1);
	farba[0] = 0.5;
	farba[1] = 0.5;
	farba[2] = 0.2;
	farba[3] = 0;
	glLightfv(GL_LIGHT0, GL_POSITION, farba);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);

	farba[0] = 0;
	farba[1] = 0;
	farba[2] = 0.2;
	farba[3] = 1;
	glMaterialfv(GL_FRONT, GL_AMBIENT, farba);
	farba[0] = 0;
	farba[1] = 0.2;
	farba[2] = 0;
	farba[3] = 1;
	glMaterialfv(GL_BACK, GL_AMBIENT, farba);

	farba[0] = 0;
	farba[1] = 0;
	farba[2] = 0;
	farba[3] = 0;
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, farba);

	farba[0] = 0;
	farba[1] = 0;
	farba[2] = 1;
	farba[3] = 1;
	glMaterialfv(GL_FRONT, GL_DIFFUSE, farba);
	farba[0] = 0;
	farba[1] = 1;
	farba[2] = 0;
	farba[3] = 1;
	glMaterialfv(GL_BACK, GL_DIFFUSE, farba);
}

void TfMain::updateViewMatrix() {
	double phi = fAzimuth / 180 * M_PI;
	double psi = fElevation / 180 * M_PI;
	glLoadIdentity();
	glTranslatef(0.5, 0.5, 0);
	gluLookAt(cos(psi) * cos(phi) * 2, cos(psi) * sin(phi) * 2, sin(psi) * 2,
		0, 0, 0, 0, 0, 1);
	glScalef(fZoom, fZoom, fZoom);
	glTranslatef(-0.5, -0.5, 0);
}

void TfMain::updateActiveVertex(int x, int y) {
	GLdouble xr, yr, zr;
	GLdouble xr2, yr2, zr2;
	GLdouble mvmatrix[16], pmatrix[16];
	GLint viewport[4];

	y = fRenderPanel->Height - y;
	glGetDoublev(GL_MODELVIEW_MATRIX, mvmatrix);
	glGetDoublev(GL_PROJECTION_MATRIX, pmatrix);
	glGetIntegerv(GL_VIEWPORT, viewport);
	gluUnProject(x, y, 0, mvmatrix, pmatrix, viewport, &xr, &yr, &zr);
	gluUnProject(x, y, 0.1, mvmatrix, pmatrix, viewport, &xr2, &yr2, &zr2);

	fActiveVertex = fPatch->getVertexNear(TVector4d(xr, yr, zr),
		TVector4d(xr2-xr, yr2-yr, zr2-zr));
}

void TfMain::drawSurface() {
	TMouseButton mouse;

	glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

	if ((fMouseDown == mbLeft) && !fActiveVertex.isValid())
		mouse = mbMiddle;
	else
		mouse = fMouseDown;

	switch (mouse) {
		case mbLeft:
			fPatch->paint(fActiveVertex, fHighlight,
				TPatch::dmWireFrame | TPatch::dmVertices | TPatch::dmGrid);
		break;
		case mbRight:
			fPatch->paint(fActiveVertex, fHighlight,
				TPatch::dmSolid | TPatch::dmHighlight | TPatch::dmIsoLines);
		break;
		default:
			fPatch->paint(fActiveVertex, fHighlight,
				TPatch::dmWireFrame | TPatch::dmSolid | TPatch::dmVertices | TPatch::dmHighlight | TPatch::dmIsoLines);
	}

	if (cbAxes->Checked) {
		glPushMatrix();
		glTranslatef(0.5, 0.5, 0);

		glDisable(GL_LIGHTING);
		glBegin(GL_LINES);
			glColor3f(1, 0, 0);
			glVertex3f(0, 0, 0);
			glVertex3f(1, 0, 0);

			glColor3f(0, 1, 0);
			glVertex3f(0, 0, 0);
			glVertex3f(0, 1, 0);

			glColor3f(0, 0, 1);
			glVertex3f(0, 0, 0);
			glVertex3f(0, 0, 1);
		glEnd();
		glEnable(GL_LIGHTING);

		glPopMatrix();
	}

	SwapBuffers(fDevCont);
}

void __fastcall TfMain::RenderPanelMouseDown(TObject *Sender,
	TMouseButton Button, TShiftState Shift, int X, int Y) {

	TVector4d v;

	fMouseDown = Button;
	fMousePos = Point(X, Y);
	switch (Button) {
		case mbLeft:
			if (Shift.Contains(ssCtrl))
				Screen->Cursor = fCursorZ;
			else
				Screen->Cursor = fCursorXY;
		break;
		case mbRight:
			Screen->Cursor = crSize;
		break;
		case mbMiddle:
			Screen->Cursor = crDefault;
			if (fActiveVertex.isValid()) {
				fWeight->eWeight->Text = FloatToStrF(
					fPatch->vertex(fActiveVertex).w(), ffGeneral, 1, 3);
				if (fWeight->ShowModal() == mrOk) {
					try {
						v = fPatch->vertex(fActiveVertex);
						v.setW(StrToFloat(fWeight->eWeight->Text));
						fPatch->setVertex(fActiveVertex, v);
					} catch (Exception &E) {
						ShowMessage(E.Message);
					}
				}
			}
		break;
		default:
			Screen->Cursor = crDefault;
	}

	if (Button == mbLeft) {
		updateActiveVertex(X, Y);
		drawSurface();
	}
}

void __fastcall TfMain::RenderPanelMouseMove(TObject *Sender, TShiftState Shift,
	int X, int Y) {

	TPatch::TIndex oldactive;
	double r;
	TVector4d v;

	switch (fMouseDown) {
		case mbLeft:
			if (fActiveVertex.isValid()) {
				if (Y < 2)
					Y = 2;
				if (Y > fRenderPanel->Height - 3)
					Y = fRenderPanel->Height - 3;
				if (X < 2)
					X = 2;
				if (X > fRenderPanel->Width - 3)
					X = fRenderPanel->Width - 3;

				if (Shift.Contains(ssCtrl)) {
					Screen->Cursor = fCursorZ;
					r = abs(sin((90 - fElevation) / 180 * M_PI));
					v = fPatch->vertex(fActiveVertex);
					v.setZ(v.z() + (double)(fMousePos.y - Y) /
						fRenderPanel->Height / r / fZoom);
					fPatch->setVertex(fActiveVertex, v);
					fMousePos = Point(X, Y);
				} else {
					Screen->Cursor = fCursorXY;
					v = fPatch->vertex(fActiveVertex);
					v.setX(v.x() - (double)(X - fMousePos.x) /
						fRenderPanel->Width * sin(fAzimuth / 180 * M_PI) /
						fZoom);
					v.setX(v.x() + (double)(Y - fMousePos.y) /
						fRenderPanel->Height / sin(fElevation / 180 * M_PI) *
						cos(fAzimuth / 180 * M_PI) / fZoom);
					v.setY(v.y() + (double)(X - fMousePos.x) /
						fRenderPanel->Width * cos(fAzimuth / 180 * M_PI) /
						fZoom);
					v.setY(v.y() + (double)(Y - fMousePos.y) /
						fRenderPanel->Height / sin(fElevation / 180 * M_PI) *
						sin(fAzimuth / 180 * M_PI) / fZoom);
					fPatch->setVertex(fActiveVertex, v);
					fMousePos = Point(X, Y);
				}

				drawSurface();
			}
		break;
		case mbRight:
			fAzimuth = fAzimuth - X + fMousePos.x;
			fElevation = fElevation + Y - fMousePos.y;
			if (fElevation >= 90)
				fElevation = 89;
			if (fElevation <= -90)
				fElevation = -89;
			updateViewMatrix();
			fMousePos = Point(X, Y);
			drawSurface();
		break;
		case mbMiddle:
			oldactive = fActiveVertex;
			updateActiveVertex(X, Y);
			if (oldactive != fActiveVertex) {
				drawSurface();
				if (!fActiveVertex.isValid())
					lWeight->Caption = "Weight: N/A";
				else
					lWeight->Caption = "Weight: " + FloatToStrF(
						fPatch->vertex(fActiveVertex).w(), ffGeneral, 1, 3);
			}
		break;
	}
}

void __fastcall TfMain::RenderPanelMouseUp(TObject *Sender, TMouseButton Button,
	TShiftState Shift, int X, int Y) {

	fMouseDown = mbMiddle;
	Screen->Cursor = crDefault;
	drawSurface();
}

void __fastcall TfMain::RenderPanelResize(TObject *Sender) {
	glViewport(0, 0, fRenderPanel->Width, fRenderPanel->Height);
}

void __fastcall TfMain::FormMouseWheel(TObject *Sender, TShiftState Shift,
	int WheelDelta, TPoint &MousePos, bool &Handled)
{
	if (WheelDelta > 0)
		fZoom = fZoom * 0.9;
	else
		fZoom = fZoom / 0.9;

	updateViewMatrix();
	fRenderPanel->Repaint();
}
//---------------------------------------------------------------------------

void TfMain::updateControls() {
	eDivisor->Text = IntToStr(fPatch->samplesPerSegment());
	eDegreeU->Text = IntToStr(fPatch->degree(TPatch::dU));
	eDegreeV->Text = IntToStr(fPatch->degree(TPatch::dV));
	mKnotsU->Clear();
	for (unsigned int i = 0; i < fPatch->knotVectorSize(TPatch::dU); i++)
		mKnotsU->Lines->Add(FloatToStrF(fPatch->knot(TPatch::dU, i), ffGeneral, 1, 3));
	mKnotsV->Clear();
	for (unsigned int i = 0; i < fPatch->knotVectorSize(TPatch::dV); i++)
		mKnotsV->Lines->Add(FloatToStrF(fPatch->knot(TPatch::dV, i), ffGeneral, 1, 3));
	eHighlightU->Text = FloatToStrF(fHighlight.u(), ffGeneral, 1, 3);
	eHighlightV->Text = FloatToStrF(fHighlight.v(), ffGeneral, 1, 3);
	eHighlightW->Text = FloatToStrF(fHighlight.w(), ffGeneral, 1, 3);
}
//---------------------------------------------------------------------------

void __fastcall TfMain::bSetDivisorClick(TObject *Sender)
{
	try {
		fPatch->setSamplesPerSegment(StrToInt(eDivisor->Text));
		fRenderPanel->Repaint();
	} catch (Exception &E) {
		ShowMessage(E.Message);
	}

	updateControls();
}
//---------------------------------------------------------------------------

void __fastcall TfMain::bSetDegreeClick(TObject *Sender)
{
	try {
		fPatch->setDegree(TPatch::dU, StrToInt(eDegreeU->Text));
		fPatch->setDegree(TPatch::dV, StrToInt(eDegreeV->Text));
		fRenderPanel->Repaint();
	} catch (Exception &E) {
		ShowMessage(E.Message);
	}

	updateControls();
}
//---------------------------------------------------------------------------

void __fastcall TfMain::bSetKnotsUClick(TObject *Sender)
{
	try {
		if ((unsigned int)mKnotsU->Lines->Count < fPatch->knotVectorSize(TPatch::dU))
			throw Exception("Too few knots!");

		for (unsigned int i = 0; i < fPatch->knotVectorSize(TPatch::dU); i++)
			fPatch->setKnot(TPatch::dU, i, StrToFloat(mKnotsU->Lines->Strings[i + 1]));

		fRenderPanel->Repaint();
	} catch (Exception &E) {
		ShowMessage(E.Message);
	}

	updateControls();
}
//---------------------------------------------------------------------------


void __fastcall TfMain::bSetKnotsVClick(TObject *Sender)
{
	try {
		if ((unsigned int)mKnotsV->Lines->Count < fPatch->knotVectorSize(TPatch::dV))
			throw Exception("Too few knots!");

		for (unsigned int i = 0; i < fPatch->knotVectorSize(TPatch::dV); i++)
			fPatch->setKnot(TPatch::dV, i, StrToFloat(mKnotsV->Lines->Strings[i + 1]));

		fRenderPanel->Repaint();
	} catch (Exception &E) {
		ShowMessage(E.Message);
	}

	updateControls();
}
//---------------------------------------------------------------------------

void __fastcall TfMain::cbAxesClick(TObject *Sender)
{
	drawSurface();
}
//---------------------------------------------------------------------------



void __fastcall TfMain::bSetHighLightClick(TObject *Sender)
{
	drawSurface();
}
//---------------------------------------------------------------------------

void __fastcall TfMain::iHighlightMouseDown(TObject *Sender,
	TMouseButton Button, TShiftState Shift, int X, int Y)
{
	fHighlight = Button;
	iHighlightMouseMove(Sender, Shift, X, Y);
}
//---------------------------------------------------------------------------

void __fastcall TfMain::iHighlightMouseMove(TObject *Sender,
	TShiftState Shift, int X, int Y)
{
	// TODO
}
//---------------------------------------------------------------------------

void __fastcall TfMain::iHighlightMouseUp(TObject *Sender,
	TMouseButton Button, TShiftState Shift, int X, int Y)
{
	Button = mbMiddle;
}
//---------------------------------------------------------------------------


