#pragma hdrstop

#include "patch.h"
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <cassert>

//---------------------------------------------------------------------------

#pragma package(smart_init)

///////////////////////////// TPatch::TIndex ///////////////////////////////////

TPatch::TIndex::TIndex(bool valid, unsigned int index) :
	fValid(valid), fIndex(index)
{
	// no code
}

bool TPatch::TIndex::isValid() const {
	return fValid;
}

unsigned int TPatch::TIndex::index() const {
	return fIndex;
}

bool TPatch::TIndex::operator==(TIndex &copy) const {
	return !(isValid() || copy.isValid()) ||
		((index() == copy.index()) && isValid() && copy.isValid());
}

bool TPatch::TIndex::operator!=(TIndex &copy) const {
	return !(*this == copy);
}

///////////////////////////// TPatch::TParam ///////////////////////////////////

TPatch::TParam::TParam(double u, double v) :
	fU(u), fV(v), fW(1.0 - u - v)
{
	// no code
}

double TPatch::TParam::u() const {
	return fU;
}

double TPatch::TParam::v() const {
	return fV;
}

double TPatch::TParam::w() const {
	return fW;
}

void TPatch::TParam::setU(double value) {
	fU = value;
	fW = 1.0 - fU - fV;
}

void TPatch::TParam::setV(double value) {
	fV = value;
	fW = 1.0 - fU - fV;
}

void TPatch::TParam::setW(double value) {
	fW = value;
	fV = 1.0 - fU - fW;
}

///////////////////////////////// TPatch ///////////////////////////////////////

TPatch::TPatch() :
	fSamplesPerSegment(20)
{
	// no code
}

TPatch::~TPatch() {
	// no code
}

unsigned int TPatch::samplesPerSegment() const {
	return fSamplesPerSegment;
}

void TPatch::setSamplesPerSegment(unsigned int value) {
	if (value != fSamplesPerSegment) {
		unsigned int oldValue = fSamplesPerSegment;
		fSamplesPerSegment = value;

		samplesPerSegmentChanged(oldValue, value);
	}
}

void TPatch::samplesPerSegmentChanged(unsigned int oldValue,
	unsigned int newValue)
{
	// NOTE - no code. Override to handle changes in number of samples.
}


