#ifndef PLOCHY_TYPES_H
#define PLOCHY_TYPES_H

template <typename T>
class TVector4 {
private:
	T fX, fY, fZ, fW;
public:
	TVector4<T>(T x = T(), T y = T(), T z = T(), T w = T());
	T x() const;
	T y() const;
	T z() const;
	T w() const;
	void setX(T value);
	void setY(T value);
	void setZ(T value);
	void setW(T value);
	TVector4<T> operator+(const TVector4<T>& copy);
	TVector4<T> operator+=(const TVector4<T>& copy);
	TVector4<T> operator-(const TVector4<T>& copy);
	TVector4<T> operator-=(const TVector4<T>& copy);
	TVector4<T> operator*(T value);
	TVector4<T> operator/(T value);
};

template <typename T>
TVector4<T> operator+(const TVector4<T>& first, const TVector4<T>& second);

template <typename T>
TVector4<T> operator-(const TVector4<T>& value);

typedef TVector4<double> TVector4d;

/// Square value.
double sqr(double value);
/// Absolute value.
double abs(double value);

//////////////////////////// implementation ////////////////////////////////////

/////////////////////////////// TVector4 ///////////////////////////////////////

template <typename T>
TVector4<T>::TVector4(T x, T y, T z, T w) :
	fX(x), fY(y), fZ(z), fW(w)
{
	// no code
}

template <typename T>
T TVector4<T>::x() const {
	return fX;
}

template <typename T>
T TVector4<T>::y() const {
	return fY;
}

template <typename T>
T TVector4<T>::z() const {
	return fZ;
}

template <typename T>
T TVector4<T>::w() const {
	return fW;
}

template <typename T>
void TVector4<T>::setX(T value) {
	fX = value;
}

template <typename T>
void TVector4<T>::setY(T value) {
	fY = value;
}

template <typename T>
void TVector4<T>::setZ(T value) {
	fZ = value;
}

template <typename T>
void TVector4<T>::setW(T value) {
	fW = value;
}

template <typename T>
TVector4<T> TVector4<T>::operator+(const TVector4<T> &copy) {
	return TVector4<T>(x() + copy.x(), y() + copy.y(), z() + copy.z(),
		w() + copy.w());
}

template <typename T>
TVector4<T> TVector4<T>::operator+=(const TVector4<T> &copy) {
	*this = *this + copy;

	return *this;
}

template <typename T>
TVector4<T> TVector4<T>::operator-(const TVector4<T> &copy) {
	return TVector4<T>(x() - copy.x(), y() - copy.y(), z() - copy.z(),
		w() - copy.w());
}

template <typename T>
TVector4<T> TVector4<T>::operator-=(const TVector4<T> &copy) {
	*this = *this - copy;

	return *this;
}

template <typename T>
TVector4<T> TVector4<T>::operator*(T value) {
	return TVector4<T>(fX * value, fY * value, fZ * value, fW * value);
}

template <typename T>
TVector4<T> TVector4<T>::operator/(T value) {
	return TVector4<T>(fX / value, fY / value, fZ / value, fW / value);
}

///////////////////////////////// global ///////////////////////////////////////

template <typename T>
TVector4<T> operator+(const TVector4<T>& first, const TVector4<T>& second) {
	TVector4<T> result(first);
	result += second;

	return result;
}

template <typename T>
TVector4<T> operator-(const TVector4<T>& value) {
	return TVector4<T>(-value.x(), -value.y(), -value.z(), value.w());
}


#endif
